@@ -49,6 +49,9 @@ def log_warn(*args):
 def bash(cmd):
     return subprocess.getoutput(cmd)
 
+def parse_json(data):
+    return json.loads(data.decode("utf-8"))
+
 class CAM(object):
     def __init__(self):
         self.__version__ = __version__
@@ -59,15 +62,21 @@ class CAM(object):
 
     def __del__(self):
         if hasattr(self, "worker_id"):
-            self._redis.hdel("workers", self.worker_id)
+            try:
+                self._redis.hdel("workers", self.worker_id)
+            except:
+                pass
         if hasattr(self, "running_job_id"):
-            self._remove_by_tid("running", self.running_job_id)
-            self.p.kill()
+            try:
+                self._remove_by_tid("running", self.running_job_id)
+                self.p.kill()
+            except:
+                pass
         
 
     def _remove_by_tid(self, part, tid):
         part_str = self._redis.lrange(part, 0, -1)
-        pending = [json.loads(d.decode("utf-8")) for d in part_str]
+        pending = [parse_json(d) for d in part_str]
         for i in range(len(pending)):
             if pending[i][0] == tid:
                 self._redis.lrem(part, 1, part_str[i])
@@ -75,7 +84,7 @@ class CAM(object):
 
     def _get_by_tid(self, part, tid):
         part_str = self._redis.lrange(part, 0, -1)
-        pending = [json.loads(d.decode("utf-8")) for d in part_str]
+        pending = [parse_json(d) for d in part_str]
         for i in range(len(pending)):
             if pending[i][0] == tid:
                 return pending[i]
@@ -83,10 +92,13 @@ class CAM(object):
 
     def _set_by_tid(self, part, tid, lst):
         part_str = self._redis.lrange(part, 0, -1)
-        pending = [json.loads(d.decode("utf-8")) for d in part_str]
+        pending = [parse_json(d) for d in part_str]
         for i in range(len(pending)):
             if pending[i][0] == tid:
                 self._redis.lset(part, i, lst)
+                return
+        self._redis.lpush(part, lst)
+        
 
     def _condition_parse(self, cond):
         #e.g.:
@@ -133,24 +145,28 @@ class CAM(object):
                 row_str = self._redis.rpop("pending")
                 if row_str is None:
                     continue
-                self.running_job_id, ptime, cmd, status = json.loads(row_str.decode("utf-8"))
+                self.running_job_id, ptime, cmd, status = parse_json(row_str)
                 cmd = "".join([cmdprefix, cmd, cmdsuffix])
                 self._redis.lpush("running", json.dumps([self.running_job_id, get_time(), cmd, get_host()]))
                 log_info("{0} Running task: {1}".format(self.worker_id, self.running_job_id))
                 log_info("{0} Running command: {1}".format(self.worker_id, cmd), )
                 self.p = Popen(cmd, shell=True)
                 worker_start_time = get_time()
-                self._redis.hset("workers", self.worker_id, json.dumps([worker_start_time, "Running %d"%self.running_job_id, cond, cmdprefix, cmdsuffix]))
                 while True:
+                    self._redis.hset("workers", self.worker_id, json.dumps([worker_start_time, "Running %d"%self.running_job_id, cond, cmdprefix, cmdsuffix]))
+                    tf = self._get_by_tid("running", self.running_job_id)
+                    taskinfo = taskinfo if tf is None else json.dumps(tf) 
+                    self._set_by_tid("running", self.running_job_id, taskinfo)
                     try:
                         if self._get_by_tid("running", self.running_job_id)[-1] == "KILLED":
                             self.p.kill()
-                            log_warn("Task ", self.running_job_id, " has been killed.")
+                            log_warn("Task ", str(self.running_job_id), " has been killed.")
                             break
                     except:
-                        if not hasattr(self, "server_disconnected"):
-                            log_warn("Server Disconnected.")
-                            self.server_disconnected = True
+                        log_warn("Server Disconnected.")
+                        #if not hasattr(self, "server_disconnected"):
+                        #    log_warn("Server Disconnected.")
+                        #    self.server_disconnected = True
                     try:
                         self.p.wait(timeout = 10)
                         break
@@ -170,6 +186,7 @@ class CAM(object):
             self._redis.lpush("pending", json.dumps([cnt, get_time(), cmd, "Pending"]))
         else:
             self._redis.rpush("pending", json.dumps([cnt, get_time(), cmd, "Pending"]))
+        log_info("New Task: ", str(cnt))
         self._redis.set('jobid', cnt + 1)
 
     def ls(self, type = None, maxwidth = None):
@@ -179,7 +196,7 @@ class CAM(object):
         <br>`cam ls worker 30` will list all workers wile each column has at most 30 chars.
         <br>`cam ls task 30` will list all tasks wile each column has at most 30 chars.
         """
-        now = get_time()
+        now = datetime.datetime.utcnow()
         log_info("Server: ", self._conf['server'], ":", str(self._conf['port']), ' v', self.__version__)
         if type is None or type == "task":
             pending = self._redis.lrange("pending", 0, -1)
@@ -187,7 +204,7 @@ class CAM(object):
             res = pending + running
             nres = []
             for i in range(len(res)):
-                data = json.loads(res[i].decode("utf-8"))
+                data = parse_json(res[i])
                 st = datetime.datetime.fromisoformat(data[1])
                 data[1] = str(now - st).split('.')[0]
                 if maxwidth is not None:
@@ -199,7 +216,7 @@ class CAM(object):
             workers = self._redis.hgetall("workers")
             info = []
             for w in workers:
-                dt = json.loads(workers[w].decode("utf-8"))
+                dt = parse_json(workers[w])
                 st = datetime.datetime.fromisoformat(dt[0])
                 lst = [w, str(now - st).split('.')[0]] + dt[1:]
                 if maxwidth is not None and len(lst) > 3:
@@ -226,6 +243,16 @@ class CAM(object):
             rrow[-1] = "KILLED"
             self._set_by_tid("running", rid, json.dumps(rrow))
 
+    def refresh(self, type = None):
+        """
+        Refresh worker list. e.g. 
+        <br>`cam refresh`
+        """
+        if type is None or type == "worker":
+            self._redis.delete("workers")
+        if type is None or type == "task":
+            self._redis.delete("running")
+
 def main():
     Cam = CAM()
     #fire.core.Display = lambda lines, out: print(*lines, file=out)
